﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SLinq;
using System.Text;

public class RoadGenerator : MonoBehaviour 
{
	public float 	smoothing = 1f;
	public List<Vector3> points = new List<Vector3>();
	
	public int roadDetail = 40;
	public Material roadMaterial;
	public Vector3[] roadLayout = new [] {
		Vector3.left * 0.5f, 
		Vector3.up * 0.1f + Vector3.left * 0.3f, 
		Vector3.up * 0.1f + Vector3.left * -0.3f, 
		Vector3.left * -0.5f
	};
	
	public void AddWorldPosition(Vector3 worldPosition)
	{
		points.Add(transform.InverseTransformPoint(worldPosition));
	}
	
	public Vector3[] GetParametizedPoints(int pointCount)
	{
		int detailCount = pointCount;
		Vector3 previousPoint;
		float length = 0f;
		float calcLength = 0f;
		float[] lengthAt = new float[detailCount];
		for (int i = 0; i < detailCount; i++)
		{
			Vector3 point = GetPosition(i / (float) (detailCount));
			if (i > 0)
				calcLength += (point - previousPoint).magnitude;
			if (i == detailCount - 1)
				length = calcLength;
			lengthAt[i] = calcLength;
			previousPoint = point;
		}
		
		Vector3[] pPoints = new Vector3[pointCount];
		int 	pointer = 0;
		float	cProgress = 0f;
		for (int i = 0; i < detailCount - 1; i++) 
		{
			int ni = i + 1;
			
			float 	lProgress1 = lengthAt[i]	/ length,
					lProgress2 = lengthAt[ni] 	/ length;
			
			float	oProgress1 = i	/ (float) (detailCount - 1),
					oProgress2 = ni	/ (float) (detailCount - 1);
			while (cProgress < lProgress2 && pointer < pointCount)
			{
				float 	lerpPosition 	= Mathf.InverseLerp(lProgress1, lProgress2, cProgress),
						actualProgress 	= Mathf.Lerp(oProgress1, oProgress2, lerpPosition);
				
				pPoints[pointer] = GetPosition(actualProgress);
				
				pointer++;
				cProgress = pointer / (float) (pointCount);
			}
		}
		return pPoints.Select(p => transform.TransformPoint(p)).ToArray();
	}
	
	public Vector3 GetTanget(int index, bool to)
	{
		int prevIndex = PreviousIndex(index),
			nextIndex = NextIndex(index);
			
		Vector3 direction = (points[nextIndex] - points[prevIndex]).normalized;
		float tangetMultiplier = 1f;
		if (to)
		{
			tangetMultiplier = -smoothing;
			tangetMultiplier *= (points[prevIndex] - points[index]).magnitude / 3;
		}
		else
		{
			tangetMultiplier = smoothing;
			tangetMultiplier *= (points[nextIndex] - points[index]).magnitude / 3;
		}
		return direction * tangetMultiplier;
	}
	
	public Vector3 GetPosition(float progress)
	{
		progress = Mathf.Clamp01(progress);
		
		int index = IndexAt(progress),
			nextIndex = NextIndex(index);
		
		float subProgress = Mathf.InverseLerp(ProgressAt(index), ProgressAt(index + 1), progress);
		
		return CalcBezier(points[index], points[nextIndex], GetTanget(index, false), GetTanget(nextIndex, true), subProgress);
	}
			
	public Vector3 CalcBezier(Vector3 p1, Vector3 p2, Vector3 t1, Vector3 t2, float progress)
	{
		Vector3 l1 = Vector3.Lerp(p1, p1 + t1		, progress),
				l2 = Vector3.Lerp(p1 + t1, p2 + t2	, progress),
				l3 = Vector3.Lerp(p2 + t2, p2		, progress);
		
		return Vector3.Lerp(Vector3.Lerp(l1, l2, progress), Vector3.Lerp(l2, l3, progress), progress);
	}
	
	protected void OnDrawGizmos()
	{
		const int detail = 50;
		Gizmos.color = Color.black;
		Gizmos.matrix = transform.localToWorldMatrix;
		
		int total = points.Count * detail;
		Vector3 previousPoint = points.FirstOrDefault(),
				point;
		for (int i = 0; i < total; i++)
		{
			float progress = i / (float) total;
			
			point = GetPosition(progress);
			Gizmos.DrawLine(previousPoint, point);
			Gizmos.DrawWireSphere(point, 0.1f);
			
			previousPoint = point;
		}
	}
	
	protected void OnDrawGizmosSelected()
	{
		Gizmos.matrix = transform.localToWorldMatrix;
		for (int i = 0; i < points.Count; i++) 
		{
			Vector3 p = points[i];
			
			Gizmos.color = Color.red;
			Gizmos.DrawSphere(p, 0.2f);
			
			Vector3 tangetTo 	= p + GetTanget(i, true),
					tangetFrom 	= p + GetTanget(i, false);
			Gizmos.color = Color.blue;
			Gizmos.DrawLine(p, tangetTo);
			Gizmos.DrawLine(p, tangetFrom);
			Gizmos.DrawSphere(tangetTo, 0.15f);
			Gizmos.DrawSphere(tangetFrom, 0.15f);
		}
	}
	
	public float ProgressAt(int index)
	{
		return (index / (float) points.Count);
	}
	
	public int IndexAt(float progress)
	{
		return Mathf.FloorToInt (points.Count * progress) % points.Count;
	}
	
	public int NextIndex(int index)
	{
		return (index + 1) % points.Count;
	}
	
	public int PreviousIndex(int index)
	{
		return (index - 1 + points.Count) % points.Count;
	}
}
